import numpy as np
import dislib as ds
from pycompss.api.task import task
from pycompss.api.constraint import constraint
from pycompss.api.reduction import reduction
from dislib.data.array import Array
from pycompss.api.parameter import COLLECTION_IN, COLLECTION_OUT, COLLECTION_INOUT, Depth, Type
from pycompss.api.api import compss_wait_on, compss_barrier
from time import time
from dislib.decomposition import tsqr
import math


@constraint(computing_units="16")
@task(
        Ur_blocks={Type: COLLECTION_OUT, Depth: 2},
        S_blocks={Type: COLLECTION_OUT, Depth: 2},
        Vr_blocks={Type: COLLECTION_OUT, Depth: 2},
    )
def _my_svd(r_block, Ur_blocks, S_blocks, Vr_blocks):
    u,s,v = np.linalg.svd(r_block, full_matrices=False)
    s = np.diag(s)
    v = v.T
    
    for i, row in enumerate(np.array_split(u, len(Ur_blocks))):
        for j, val in enumerate(np.array_split(row, len(Ur_blocks[i]), axis=1)):
            Ur_blocks[i][j] = np.asfortranarray(val)

    for i, row in enumerate(np.array_split(s, len(S_blocks))):
        for j, val in enumerate(np.array_split(row, len(S_blocks[i]), axis=1)):
            S_blocks[i][j] = np.asfortranarray(val)

    for i, row in enumerate(np.array_split(v, len(Vr_blocks))):
        for j, val in enumerate(np.array_split(row, len(Vr_blocks[i]), axis=1)):
            Vr_blocks[i][j] = np.asfortranarray(val)


def check_convergeces(res, tol, nsv):
    res = np.block(res).flatten()
    converged = res < tol
    return np.all(converged[:nsv]), max(res[:nsv])


@constraint(computing_units="4")
@task(B_blocks={Type: COLLECTION_INOUT, Depth: 2}, returns=1)
def R_ortho(B_blocks, R, k , b, B_reg_shape):
    B = np.block(B_blocks)
    R = R.astype(np.float64)
    
    for j in range(0, k, b):
        w = B[:,j:j+b]
        if j > 0:
            h = B[:,0:j].T @ w
            w = w - B[:,0:j] @ h
        w, r1 = np.linalg.qr(w)

        if j > 0:
            h2 = B[:,0:j].T @ w
            w = w - B[:,0:j] @ h2
            R[0:j,j:j+b][:] = h + h2
        w, r2 = np.linalg.qr(w)

        B[:,j:j+b] = w
        R[j:j+b,j:j+b][:] = r1 @ r2
        
    n, m = B_reg_shape
    for i in range(len(B_blocks)):
        for j in range(len(B_blocks[i])):
            B_blocks[i][j] = np.asfortranarray(B[i*n:(i+1)*n, j*m:(j+1)*m])
    
    return R


def _get_all_cols(A, i):
    blocks = [[] for j in range(A._n_blocks[0])]
    shape = [A.shape[0], 0]
    for col_idx in range(i+1):
        col_shape = A._get_col_shape(col_idx)
        shape[1] += col_shape[1]
        for j in range(A._n_blocks[0]):
            blocks[j].append(A._blocks[j][i])
    return Array(blocks=blocks,
                    top_left_shape=A._reg_shape,
                    reg_shape=A._reg_shape, shape=col_shape,
                    sparse=A._sparse, delete=False)

@task()
def _norm(b):
    return np.linalg.norm(b, axis=0)

@reduction(chunk_size="8")
@task(blocks={Type: COLLECTION_IN, Depth: 1})
def _norm_red(blocks):
    return np.linalg.norm(blocks, axis=0)

def my_norm(A):
    norm_blocks = Array._get_out_blocks(A._n_blocks)
    for i in range(A._n_blocks[0]):
        for j in range(A._n_blocks[1]):
            norm_blocks[i][j] = _norm(A._blocks[i][j])

    norms = []
    for j in range(A._n_blocks[1]):
        col = [norm_blocks[i][j] for i in range(A._n_blocks[0])]
        if len(col) > 1:
            norms.append(_norm_red(col))
        else:
            norms.append(col[0])
    return norms


def nsv_tolerance(m, n, nsv, S):
	min_sval = S[nsv-1]
	total_sval = min(m, n)
	sum_nsv = sum([s**2 for s in S])
	sum_remaining = (total_sval-nsv) * min_sval**2
	tol = np.sqrt(sum_remaining / (sum_remaining + sum_nsv))
	return tol
    

def _svd_random(A, b, k, max_restarts, nsv, tol, verbose=False):
    # A: matrix
    # b: block size
    # k: number of vectors (must be multiple of b)
    # max_restarts: maximum number of restarts
    # nsv: number of desired singular values
    # tol: consider a singular converged when its residual is less than tol
    m, n = A.shape

    arity = 2
    cus = 4
    
    B = ds.random_array((n, k), (b, b), random_state=0)
    R = ds.zeros((k, k), (k, k))
    # print('************ Q = A @ B **************')
    # print(A.shape, A._reg_shape)
    # print(B.shape, B._reg_shape)
    # print(arity)
    # print('***********************')
    Q = ds.matmul(A, B)
    # compss_barrier()

    for iters in range(max_restarts):
        for j in range(0, k, b):
            col_idx = j // b
            q = Q._get_col_block(col_idx)
            if j > 0:
                h = ds.matmul(_get_all_cols(Q, col_idx), q, transpose_a=True)
                q = ds.data.matsubtract(q, ds.matmul(_get_all_cols(Q, col_idx), h))
                h = ds.matmul(_get_all_cols(Q, col_idx), q, transpose_a=True)
                q = ds.data.matsubtract(q, ds.matmul(_get_all_cols(Q, col_idx), h))

            # print('************** TSQR')
            q, _ = tsqr(q, mode='reduced')
            q._delete = False
            
            for i in range(Q._n_blocks[0]):
                Q._blocks[i][col_idx] = q._blocks[i][0]
            

        # print('************ B = A @ Q **************')
        # print(A.shape, A._reg_shape)
        # print(Q.shape, Q._reg_shape)
        # print(arity)
        # print('***********************')
        # B = matmul_fma(A, Q, transpose_a=True, arity=arity, cu=16) #
        B = ds.matmul(A, Q, transpose_a=True)
        # compss_barrier()

        R._blocks[0][0] = R_ortho(B._blocks, R._blocks[0][0], k, b, B._reg_shape)
        
        Ur_blocks = Array._get_out_blocks((k//b, k//b))
        Vr_blocks = Array._get_out_blocks((k//b, k//b))
        S_blocks = Array._get_out_blocks((k//b, k//b))
        
        _my_svd(R._blocks[0][0], Ur_blocks, S_blocks, Vr_blocks) 

        Ur = Array(Ur_blocks, (b, b), (b, b), (R.shape[0], R.shape[0]), False)
        S = Array(S_blocks, (b, b), (b, b), (R.shape[1], R.shape[1]), False)
        Vr = Array(Vr_blocks, (b, b), (b, b), (R.shape[1], R.shape[1]), False)

        # print('************ U = Q @ Vr **************')
        # print(Q.shape, Q._reg_shape)
        # print(Vr.shape, Vr._reg_shape)
        # print(arity)
        # print('***********************')
        U = ds.matmul(Q, Vr)
        # compss_barrier()

        # print('************ V = B @ Ur **************')
        # print(B.shape, B._reg_shape)
        # print(Ur.shape, Ur._reg_shape)
        # print(arity)
        # print('***********************')
        V = ds.matmul(B, Ur)
        # compss_barrier()

        # print('************ Q = A @ V **************')
        # print(A.shape, A._reg_shape)
        # print(V.shape, V._reg_shape)
        # print(arity)
        # print('***********************')
        Q = ds.matmul(A, V)
        # compss_barrier()


        # print('************ U @ S **************')
        # print(U.shape, U._reg_shape)
        # print(S.shape, S._reg_shape)
        # print(arity)
        # print('***********************')
        mmm = ds.data.matsubtract(Q, ds.matmul(U, S))

        res = np.block(compss_wait_on(my_norm(mmm)))
        converged, conv_tol = check_convergeces(res, tol, nsv)
        if converged:
            if verbose:
                print("Converged in {} iterations to {} precision.".format(iters+1, conv_tol))
            break
        else:
            if verbose:
                print("Not converged in {} iterations. Current precision: {}.".format(iters+1, conv_tol))

    return U, S, V, conv_tol


def random_svd(A, iters, epsilon, tol, nsv=None, k=None, verbose=False):
    """ Random SVD

    Parameters
    ----------
    a : ds-arrays
        Input ds-array.
    iters : int
        Number of inner iterations for the Random algorithm to converge.
    bs : int
        Block size (in the column axis)
    k : int
        Number of restarting vectors (must be a multiple of bs)
    nsv : int
        Number of desired singular values
    tolerance : float64
        If the residual value of a singular value is less than the
        tolerante, that singular value is considered to be converged.
    epsilon : float64
        Value that defines the number of singular values required,
        as it is reduced, the number of singular values required
        is increased.
    verbose : bool
        Controls the verbosity for the algorithm convergence. Shows convergence accuracy
        and singular value criteria for each iteration.

    Returns
        -------
        U : ds-array
            The U of the matrix, Unitary array returned as ds-array, the shape
            is A.shape[0] x rank, and the block size is the block size of
            A in the row axis x bs.
        S : ds-array
            The S of the matrix. It is represented as a 2-dimensional matrix,
            the diagonal of this matrix is the vector with the singular
            values. Its shape is rank x rank and the block size is bs x bs
        V : ds-array
            The V of the matrix, Unitary array returned as ds-array,
            the shape is A.shape[1] x rank, and the block size is bs x bs

        Raises
        ------
        ValueError
            If num_sv is bigger than the number of columns
            or
            If k < num_nsv
            or
            If k % b != 0
    """
    m, n = A.shape
    b = A._reg_shape[1]

    if nsv is None:
        nsv = np.argmin((np.logspace(0, -20, n) - epsilon)**2)

    if k is None:
        if b == nsv:
            k = b
        elif b > nsv:
            k = b
        else:
            k = math.ceil(nsv / b) * b

    
    if nsv > n:
        raise ValueError("Number of singular values to compute can't"
                         " be bigger than the total number of singular"
                         "values of the matrix.")
    if k < nsv:
        raise ValueError("Rank should be at least the number of singular"
                         "values to compute.")

    if k % b != 0:
        raise ValueError("K should be multiple of `A` block size.")

    if verbose:
        print(f'Random-SVD: Starting computing {nsv} singular values using {k} vectors.')

    epsilon_conv = False
    while not epsilon_conv:
        U, S, V, conv_tol = _svd_random(A, b, k, iters, nsv, tol, verbose=verbose)

        if conv_tol > tol:
            print(f'WARNING: Random-SVD did not converge in {iters} iterations to {tol} precision.')
            print(f'         Maximum accuracy was {conv_tol}')
        
        S_flat = np.diag(S.collect())
        epsilon_hat = nsv_tolerance(m, n, nsv, S_flat)
        if epsilon_hat <= epsilon:
            epsilon_conv = True
        else:
            k += b
            nsv += b
            if verbose:
                print(f'NOTE: Random-SVD singular value criteria {epsilon_hat} > {epsilon}.')
                print(f'      Increasing number of vectors to {k} and {nsv} singular values.')

    if verbose:
        print(f'Random-SVD: Converged to {epsilon_hat} singular value tolerance.')
        
    return U, S, V.T, nsv

@task(returns=1, is_distributed=True)
def load_npy(file_path):
    return np.load(file_path)


def load_dataset(wildcard, vertical):
    from glob import glob
    import re, os

    files = glob(wildcard)
    if vertical:
        files = sorted(files, key=lambda x: int(re.findall(r'\d+', os.path.basename(x))[0]))
    else:
        files = sorted(files, key=lambda x: int(''.join(re.findall(r'\d+', os.path.basename(x)))))

    if vertical:
        blocks = []
    else:
        blocks = [[]]
    shape = [0, 0]
    bs = None

    for f in files:
        with open(files[0], "rb") as fid:
            version = np.lib.format.read_magic(fid)
            shape_block, _, _ = np.lib.format._read_array_header(fid, version)
            if not bs:
                bs = shape_block
            idx = 1 if vertical else 0
            if shape[idx] == 0:
                shape[idx] = shape_block[idx]
            idx = 0 if vertical else 1
            shape[idx] += shape_block[idx]
            if vertical:
                blocks[0].append(load_npy(f))
            else:
                blocks.append([load_npy(f)])

    return Array(blocks, bs, bs, shape, False)


def load_M1():
    A = load_dataset('/gpfs/projects/bsc19/bsc19756/MakeTrainingMatricesSerial/snapshots_at_step_*', vertical=True)
    A = A.rechunk((75_000, 500))

def load_M2():
    A = load_dataset('/gpfs/projects/bsc19/bsc19756/Matrix_Sebastian/*', vertical=False)
    A = A.rechunk((40_000, 1450))


def load_new_dataset(path):
    import os

    files = os.listdir(path)

    blocks = [[]]
    shape = [0, 0]
    bs = None

    files = [os.path.join(path, f) for f in files]

    for f in files:
        with open(f, "rb") as fid:
            version = np.lib.format.read_magic(fid)
            shape_block, _, _ = np.lib.format._read_array_header(fid, version)
            if not bs:
                bs = shape_block
            idx = 0
            if shape[idx] == 0:
                shape[idx] = shape_block[idx]
            idx = 1 
            shape[idx] += shape_block[idx]
            blocks[0].append(load_npy(f))

    return Array(blocks, bs, bs, shape, False)

if __name__ == "__main__":
    import sys 

    path = sys.argv[1]

    A = load_new_dataset(path)
    compss_barrier()
    
    t1 = time()
    print(A.shape, A._reg_shape)

    A = A.rechunk((50_000, 456))
    # compss_barrier()
    U, S, V, nsv = random_svd(A, iters=1, nsv=456, tol=1e-3, epsilon=1e-3)

    # compss_barrier()

    duration = time()-t1
    print(f'Time R-SVD:', duration)
    print(np.diag(S.collect())[:nsv])
